var registry = {};
var seen = {};
var config = {};
var defaultConfig = {
    relativePaths: true,
    defaultFile: 'main',
    paths: null
};

var define = function(name, deps, callback) {

    var value = {};

    if (!callback) {

        value.deps = [];
        value.callback = deps;

    } else {

        value.deps = deps;
        value.callback = callback;

    }

    if(registry[name]){
        throwError('duplicatedModule ' + name);
    }

    registry[name] = value;

};

var require = function(name, submodule) {

    var module = internalRequire(name, null);
    // when module is undefined, accessing a property like 'default'
    // triggers a type error
    // appears when is added custom error method with .on('error', func)
    // add tests for that
    var result = module[submodule || 'default'];

    return cloneObject(result);

};

function internalRequire(name, referrerName) {

    name = getNameWithPrefix(name);
    var exports = seen[name];
    var oldName = name;

    if (exports !== undefined) {
        return exports;
    }

    if (!registry[name]) {

        var hasEndingSlash = name.slice(-1) === '/';
        var slash = hasEndingSlash ? '' : '/';
        var indexFileName = name + slash + config.defaultFile;

        if(registry[indexFileName]){

            name = indexFileName;

        } else if (referrerName) {
            throwError('Could not find module ' + name + ' required by: ' + referrerName);
        } else {
            throwError('Could not find module ' + name);
        }

    }

    exports = seen[name] = seen[oldName] = {};

    var mod = registry[name];
    var deps = mod.deps;
    var callback = mod.callback;
    var reified = [];
    var length = deps.length;

    for (var i = 0; i < length; i++) {

        if (deps[i] === 'exports') {
            reified.push(exports);
        } else {
            reified.push(internalRequire(resolve(deps[i], name), name));
        }

    }

    // test "instances of objects: should export method" exposes global variable
    // "this" sent as context in callback.apply is the cause of the pollution of globalnamespace
    // TODO fix the pollution or find a way to restrict that kind of methods by throwing errors
    callback.apply(this, reified);

    return exports;

}

function getNameWithPrefix(path){

    var prefixExitsOnConfig = config && config.paths && typeof config.paths.prefix === 'string';
    var prefix = prefixExitsOnConfig ? config.paths.prefix : '';
    var ignorePatterns = getIgnoredPatterns();
    var shouldIgnore = shouldIgnorePrefix(path, ignorePatterns);
    var shouldAddPrefix = prefixExitsOnConfig && !shouldIgnore;

    prefix = prefix.length && prefix.slice(-1) !== '/' ? prefix + '/' : prefix;

    var name = shouldAddPrefix ? prefix + path : path;

    return name;

}

function getIgnoredPatterns(){

    var prefixExitsOnConfig = config && config.paths && typeof config.paths.prefix === 'string';
    var hasIgnore = prefixExitsOnConfig && config.paths.ignore;
    var ignoreIsString = hasIgnore && typeof config.paths.ignore === 'string';
    var ignorePatterns = [];

    if(hasIgnore){

        if(ignoreIsString){
            ignorePatterns.push(config.paths.ignore);
        } else if (isArray(config.paths.ignore)){
            ignorePatterns = config.paths.ignore;
        }

    }

    return ignorePatterns;

}

function shouldIgnorePrefix(path, ignorePatterns){

    var shouldIgnore = false;

    for (var i = 0; i < ignorePatterns.length; i++) {

        var containsInvalidChars = /[^\w\d/_-]/.test(ignorePatterns[i]);

        if(containsInvalidChars){

            throwError('Ignored prefixes contains invalid chars');

        }

        var ignorePattern = new RegExp('^' + ignorePatterns[i] + '(\/)?(\w+)*');

        if(shouldIgnore === false && ignorePattern.test(path)){
            shouldIgnore = true;
        }

    };

    return shouldIgnore;

}

function resolve(child, name) {

    if (child.charAt(0) === '.') {
        if(!config.relativePaths){
            throwError('Relative paths are not allowed: ' + child + ' required by ' + name);
        }
    } else {
        return child;
    }

    var parts = child.split('/');
    var parentBase = name.split('/').slice(0, -1);

    for (var i = 0, l = parts.length; i < l; i++) {

        var part = parts[i];

        if (part === '..') {
            parentBase.pop();
        } else if (part === '.') {
            continue;
        } else {
            parentBase.push(part);
        }

    }

    return parentBase.join('/');
}

function cloneObject(source) {
    
    var name, s, empty = {}, dest = {};
    var isNotObject = !isObject(source);
    
    if(isNotObject){
        return source;
    }

    for(name in source){
        /*jshint -W089 */
        // the (!(name in empty) || empty[name] !== s) condition avoids copying properties in "source"
        // inherited from Object.prototype.  For example, if dest has a custom toString() method,
        // don't overwrite it with the toString() method that source inherited from Object.prototype
        s = source[name];
        if(!(name in dest) || (dest[name] !== s && (!(name in empty) || empty[name] !== s))){
            dest[name] = cloneObject ? cloneObject(s) : s;
        }
    }

    return dest;

}

function isArray(myArray) {
    return Object.prototype.toString.call(myArray).indexOf("Array") > -1;
}

function isObject(obj){
  return Object.prototype.toString.call(obj) === '[object Object]';
}

function isString(string){
  return Object.prototype.toString.call(string) === '[object String]';
}

function isFunction(func){
  return Object.prototype.toString.call(func) === '[object Function]';
}

function extend(obj1,obj2){
  
    var obj3 = {};
    
    if(isObject(obj1)){
      for (var attrname in obj1) { 
        obj3[attrname] = cloneObject(obj1[attrname]); 
      }
    }
    
    if(isObject(obj2)){
      for (var attrname in obj2) { 
        
        if(isObject(obj1) && isObject(obj1[attrname]) && isObject(obj2[attrname])){
          obj3[attrname] = cloneObject(extend(obj1[attrname], obj2[attrname]));
        } else {
          obj3[attrname] = obj2[attrname];  
        }
        
      }
    }
    
    return cloneObject(obj3);
    
}

function throwError(msg){

    throw new Error(msg);

}

function Config(data) {

    if(data === undefined){

        return config;

    } else if(isString(data)) {

        return config[data];

    } else if (isObject(data)){

        config = extend(defaultConfig, data);

    }

}

function listModuleNames(){

    var moduleName;
    var names = [];

    for (moduleName in registry) {
        names.push(moduleName);
    };

    return names;

}

function markAsUnseen(moduleName){
    delete seen[moduleName];
}

// make tests for reset
// when events is not reset, throw tests are failing but without being able to figure it out fast
function reset() {
    registry = {};
    seen = {};
    Config({});
}

Config({});

var AMD = {
    define: define,
    require: require,
    config: Config,
    markAsUnseen: markAsUnseen,
    list_entries: listModuleNames,
    reset: reset
};
 module.exports = AMD;